import xml
import xml.dom
from xml.dom import minidom
from LuciSyslog import LuciSyslog
from conga_ssl import SSLSocket

CERTS_DIR_PATH = '/var/lib/luci/var/certs/'

try:
    luci_log = LuciSyslog()
except:
    pass

class RicciError(Exception):
    pass

class RicciCommunicator:
    def __init__(self, hostname, enforce_trust=True, port=11111):
        self.__hostname = hostname
        self.__port = port
        
        self.__timeout_init  = 4
        self.__timeout_auth  = 4
        self.__timeout_short = 6
        self.__timeout_long  = 600
        
        self.__privkey_file = CERTS_DIR_PATH + 'privkey.pem'
        self.__cert_file = CERTS_DIR_PATH + 'cacert.pem'
        
        try:
            self.ss = SSLSocket(self.__hostname,
                                self.__port,
                                self.__timeout_init)
            if enforce_trust:
                if not self.ss.trusted():
                    raise RicciError, 'ricci\'s certificate is not trusted'
        except Exception, e:
            raise RicciError, 'Error setting up SSL for connection to %s: %s' \
                % (self.__hostname, str(e))
        except:
            raise RicciError, 'Error setting up SSL for connection to %s' \
                % self.__hostname
        
        # receive ricci header
        hello = self.__receive(self.__timeout_init)
        try:
            luci_log.debug_verbose('RC:init0: Received header from %s: \"%s\"' \
                % (self.__hostname, hello.toxml()))
        except:
            pass
        
        self.__authed = hello.firstChild.getAttribute('authenticated') == 'true'
        self.__cluname = hello.firstChild.getAttribute('clustername')
        self.__clualias = hello.firstChild.getAttribute('clusteralias')
        self.__reported_hostname = hello.firstChild.getAttribute('hostname')
        self.__os = hello.firstChild.getAttribute('os')
        self.__dom0 = hello.firstChild.getAttribute('xen_host') == 'true'
        
        pass

    def hostname(self):
        luci_log.debug_verbose('RC:hostname: [auth %d] reported hostname = %s' \
            % (self.__authed, self.__hostname))
        return self.__hostname
    def authed(self):
        luci_log.debug_verbose('RC:authed: reported authed = %d for %s' \
            % (self.__authed, self.__hostname))
        return self.__authed
    def system_name(self):
        luci_log.debug_verbose('RC:system_name: [auth %d] reported system_name = %s for %s' \
            % (self.__authed, self.__reported_hostname, self.__hostname))
        return self.__reported_hostname
    def cluster_info(self):
        luci_log.debug_verbose('RC:cluster_info: [auth %d] reported cluster_info = (%s,%s) for %s' \
            % (self.__authed, self.__cluname, self.__clualias, self.__hostname))
        return (self.__cluname, self.__clualias)
    def os(self):
        luci_log.debug_verbose('RC:os: [auth %d] reported os = %s for %s' \
            % (self.__authed, self.__os, self.__hostname))
        return self.__os
    def dom0(self):
        luci_log.debug_verbose('RC:dom0: [auth %d] reported dom0 = %s for %s' \
            % (self.__authed, self.__dom0, self.__hostname))
        return self.__dom0

    def fingerprint(self):
        return self.ss.peer_fingerprint()

    def trust(self):
        return self.ss.trust()

    def untrust(self):
        return self.ss.untrust()

    def trusted(self):
        return self.ss.trusted()

    def auth(self, password):
        if self.authed():
            luci_log.debug_verbose('RC:auth0: already authenticated to %s' \
                % self.__hostname)
            return True
        
        # send request
        doc = minidom.Document()
        ricci = doc.createElement("ricci")
        ricci.setAttribute("version", "1.0")
        ricci.setAttribute("function", "authenticate")
        ricci.setAttribute("password", password)
        doc.appendChild(ricci)
        self.__send(doc, self.__timeout_auth)
        
        # receive response
        resp = self.__receive(self.__timeout_auth)
        self.__authed = resp.firstChild.getAttribute('authenticated') == 'true'

        if self.__authed:
            try:
                self.__cluname = resp.firstChild.getAttribute('clustername')
                self.__clualias = resp.firstChild.getAttribute('clusteralias')
                self.__reported_hostname = resp.firstChild.getAttribute('hostname')
                self.__os = resp.firstChild.getAttribute('os')
                self.__dom0 = resp.firstChild.getAttribute('xen_host') == 'true'
            except:
                pass
        luci_log.debug_verbose('RC:auth1: auth call returning %d' \
            % self.__authed)
        return self.__authed

    def unauth(self):
        doc = minidom.Document()
        ricci = doc.createElement('ricci')
        ricci.setAttribute('version', '1.0')
        ricci.setAttribute('function', 'unauthenticate')
        doc.appendChild(ricci)
        self.__send(doc, self.__timeout_auth)
        resp = self.__receive(self.__timeout_auth)

        luci_log.debug_verbose('RC:unauth0: trying to unauthenticate to %s' \
            % self.__hostname)

        try:
            ret = resp.firstChild.getAttribute('success')
            luci_log.debug_verbose('RC:unauth1: unauthenticate returned %s for %s' \
                % (ret, self.__hostname))
            if ret != '0':
                raise Exception, 'Invalid response'
            try:
                self.ss.untrust()
            except:
                pass
        except:
            errstr = 'Error authenticating to host %s: %s' \
                        % (self.__hostname, str(ret))
            luci_log.debug_verbose('RC:unauth2:' + errstr)
            raise RicciError, errstr
        return True


    def process_batch(self, batch_xml, async=False):
        try:
            luci_log.debug_verbose('RC:PB0: [auth=%d] to %s for batch %s [async=%d]' \
                % (self.__authed, self.__hostname, batch_xml.toxml(), async))
        except:
            pass

        if not self.authed():
            raise RicciError, 'not authenticated to host %s' % self.__hostname
        
        # construct request
        doc = minidom.Document()
        ricci = doc.createElement("ricci")
        ricci.setAttribute("version", "1.0")
        ricci.setAttribute("function", "process_batch")
        async_str = None
        if async:
            async_str = 'true'
        else:
            async_str = 'false'
        ricci.setAttribute("async", async_str)
        doc.appendChild(ricci)
        ricci.appendChild(batch_xml.cloneNode(True))
        
        # send request
        try:
            self.__send(doc, self.__timeout_short)
        except Exception, e:
            luci_log.debug_verbose('RC:PB1: Error sending XML \"%s\" to host %s' \
                % (doc.toxml(), self.__hostname))
            raise RicciError, 'Error sending XML to host %s: %s' \
                    % (self.__hostname, str(e))
        except:
            raise RicciError, 'Error sending XML to host %s' % self.__hostname
        
        # receive response
        doc = self.__receive(self.__timeout_long)
        try:
            luci_log.debug_verbose('RC:PB2: received from %s XML \"%s\"' \
                % (self.__hostname, doc.toxml()))
        except:
            pass
 
        if doc.firstChild.getAttribute('success') != '0':
            luci_log.debug_verbose('RC:PB3: batch command failed')
            raise RicciError, 'The last ricci command to host %s failed' \
                    % self.__hostname
        
        batch_node = None
        for node in doc.firstChild.childNodes:
            if node.nodeType == xml.dom.Node.ELEMENT_NODE:
                if node.nodeName == 'batch':
                    batch_node = node.cloneNode(True)
        if batch_node == None:
            luci_log.debug_verbose('RC:PB4: batch node missing <batch/>')
            raise RicciError, 'missing <batch/> in ricci\'s response from %s' \
                    % self.__hostname

        return batch_node
    
    def batch_run(self, batch_str, async=True):
        try:
            batch_xml_str = '<?xml version="1.0" ?><batch>' + batch_str + '</batch>'
            luci_log.debug_verbose('RC:BRun0: attempting batch \"%s\" for host %s' \
                % (batch_xml_str, self.__hostname))
            batch_xml = minidom.parseString(batch_xml_str).firstChild
        except Exception, e:
            luci_log.debug_verbose('RC:BRun1: received invalid batch XML for %s: \"%s\": %s' \
                % (self.__hostname, batch_xml_str, str(e)))
            raise RicciError, 'batch XML is malformed'

        try:
            ricci_xml = self.process_batch(batch_xml, async)
            try:
                luci_log.debug_verbose('RC:BRun2: received XML \"%s\" from host %s in response to batch command.' \
                    % (ricci_xml.toxml(), self.__hostname))
            except:
                pass
        except:
            luci_log.debug_verbose('RC:BRun3: An error occurred while trying to process the batch job: \"%s\"' % batch_xml_str)
            return None

        doc = minidom.Document()
        doc.appendChild(ricci_xml)
        return doc

    def batch_report(self, batch_id):
        luci_log.debug_verbose('RC:BRep0: [auth=%d] asking for batchid# %s for host %s' \
            % (self.__authed, batch_id, self.__hostname))

        if not self.authed():
            raise RicciError, 'Not authenticated to host %s' % self.__hostname
        
        # construct request
        doc = minidom.Document()
        ricci = doc.createElement("ricci")
        ricci.setAttribute("version", "1.0")
        ricci.setAttribute("function", "batch_report")
        ricci.setAttribute("batch_id", str(batch_id))
        doc.appendChild(ricci)
        
        # send request
        self.__send(doc, self.__timeout_short)
 
       
        # receive response
        doc = self.__receive(self.__timeout_short)
        if doc.firstChild.getAttribute('success') == '12':
            return None
        if doc.firstChild.getAttribute('success') != '0':
            raise RicciError, 'Error while retrieving batch report for batch #%d from host %s' % (batch_id, self.__hostname)
        batch_node = None
        for node in doc.firstChild.childNodes:
            if node.nodeType == xml.dom.Node.ELEMENT_NODE:
                if node.nodeName == 'batch':
                    batch_node = node.cloneNode(True)
        if batch_node == None:
            raise RicciError, 'Missing <batch/> in ricci\'s response from host %s' % self.__hostname
        return batch_node
    
    
    
    
    
    def __send(self, xml_doc, timeout):
        buff = xml_doc.toxml() + '\n'
        try:
            self.ss.send(buff, timeout)
        except Exception, e:
            luci_log.debug_verbose('RC:send0: Error sending XML \"%s\" to %s: %s' \
                                   % (buff, self.__hostname, str(e)))
            raise RicciError, 'write error while sending XML to host %s' \
                  % self.__hostname
        except:
            raise RicciError, 'write error while sending XML to host %s' \
                  % self.__hostname
        try:
            luci_log.debug_verbose('RC:send1: Sent XML \"%s\" to host %s' \
                % (xml_doc.toxml(), self.__hostname))
        except:
            pass
        return
    
    def __receive(self, timeout):
        doc = None
        xml_in = ''
        try:
            xml_in = self.ss.recv(timeout)
        except Exception, e:
            luci_log.debug_verbose('RC:recv0: Error reading data from %s: %s' \
                % (self.__hostname, str(e)))
            raise RicciError, 'Error reading data from host %s' % self.__hostname
        except:
            raise RicciError, 'Error reading data from host %s' % self.__hostname
        luci_log.debug_verbose('RC:recv1: Received XML \"%s\" from host %s' \
            % (xml_in, self.__hostname))

        try:
            if doc == None:
                doc = minidom.parseString(xml_in)
        except Exception, e:
            luci_log.debug_verbose('RC:recv2: Error parsing XML \"%s" from %s' \
                % (xml_in, str(e)))
            raise RicciError, 'Error parsing XML from host %s: %s' \
                    % (self.__hostname, str(e))

        if not doc or not doc.firstChild:
            raise RicciError, \
                    'Error an empty response was received from host %s' \
                    % self.__hostname
        
        try:        
            if doc.firstChild.nodeName != 'ricci':
                luci_log.debug_verbose('RC:recv3: Expecting \"ricci\" got XML \"%s\" from %s' %
                    (xml_in, self.__hostname))
                raise Exception, 'Expecting first XML child node to be \"ricci\"'
        except Exception, e:
            raise RicciError, 'Invalid XML ricci response from host %s' \
                    % self.__hostname
        
        return doc
    


def get_ricci_communicator(self, hostname, allowed_systems):
    if not self.access_to_host_allowed(hostname, allowed_systems):
        return None
    
    try:
        return RicciCommunicator(hostname)
    except Exception, e:
        luci_log.debug_verbose('RC:GRC0: Error creating a ricci connection to %s: %s' \
            % (hostname, str(e)))
        return None
    pass


def ricci_get_called_hostname(self, ricci):
    return ricci.hostname()
def ricci_get_reported_hostname(self, ricci):
    return ricci.system_name()
def ricci_get_os(self, ricci):
    return ricci.os()
def ricci_get_dom0(self, ricci):
    return ricci.dom0()
def ricci_get_cluster_info(self, ricci):
    return ricci.cluster_info()
def ricci_get_authenticated(self, ricci):
    return ricci.authed()
def ricci_authenticate(self, ricci, password):
    return ricci.auth(password)
def ricci_unauthenticate(self, ricci):
    return ricci.unauth()








########## helpers to process batch as returned by ricci #############





# check the status of batch
# returns (int num, int total)
# * total:
#          total number of modules in batch
# * num:
#          if num == total: 
#             all modules in the batch completed successfuly
#          if num > 0: 
#             last seq. number of module that successfuly finished
#          if num < 0: 
#             module (-num) failed (next module won't be processed)
def batch_status(batch_xml):
    if batch_xml.nodeName != 'batch':
        try:
            luci_log.debug_verbose('RC:BS0: Expecting an XML batch node. Got \"%s\"' \
                % batch_xml.toxml())
        except:
            pass
        raise RicciError, 'Not an XML batch node'
    total = 0
    last  = 0
    for node in batch_xml.childNodes:
        if node.nodeType == xml.dom.Node.ELEMENT_NODE:
            if node.nodeName == 'module':
                total = total + 1
                status = node.getAttribute('status')
                if status == '0':
                    # success
                    last = last + 1
                elif status == '3' or status == '4':
                    # failure
                    last = last + 1
                    last = last - 2 * last
    try:
        luci_log.debug_verbose('RC:BS1: Returning (%d, %d) for batch_status(\"%s\")' \
            % (last, total, batch_xml.toxml()))
    except:
        luci_log.debug_verbose('RC:BS2: Returning last, total')

    return (last, total)



# extract error_code from module's response
# * module_num:
#              1-based seq. number of module to process
#
# returns (int error_code, string error_msg)
# * error_code: each module defines own error codes, which are >0
#          -101 - in progress
#          -102 - scheduled
#          -103 - removed from schedule
#          -104 - failed to execute module
# 
#          >-3  - module executed. Following codes are defined:
#             -2   - API error
#             -1   - undefined error occured (msg not necesarily very informative)
#             0    - no error (msg is empty string)
#             >0   - predefined error has occured
#                        (check respective API, msg will be fully descriptive)
# * error_msg:  error message
def extract_module_status(batch_xml, module_num=1):
    if batch_xml.nodeName != 'batch':
        luci_log.debug_verbose('RC:EMS0: Expecting \"batch\" got \"%s\"' % batch_xml.toxml())
        raise RicciError, 'Invalid XML node; expecting a batch node'

    c = 0
    for node in batch_xml.childNodes:
        if node.nodeType == xml.dom.Node.ELEMENT_NODE:
            if node.nodeName == 'module':
                module_xml = node
                c = c + 1
                if c == module_num:
                    status = module_xml.getAttribute('status')
                    if status == '0' or status == '4':
                        # module executed, dig deeper into request
                        for node_i in module_xml.childNodes:
                            if node_i.nodeType == xml.dom.Node.ELEMENT_NODE:
                                if node_i.nodeName == 'API_error':
                                    return -2, 'API error'
                                elif node_i.nodeName == 'response':
                                    for node_j in node_i.childNodes:
                                        if node_j.nodeType == xml.dom.Node.ELEMENT_NODE:
                                            if node_j.nodeName == 'function_response':
                                                code = -11111111
                                                msg  = 'BUG'
                                                for var in node_j.childNodes:
                                                    if var.nodeType == xml.dom.Node.ELEMENT_NODE:
                                                        if var.nodeName == 'var':
                                                            if var.getAttribute('name') == 'success' and var.getAttribute('value') == 'true':
                                                                return 0, ''
                                                            elif var.getAttribute('name') == 'error_code':
                                                                code = int(var.getAttribute('value'))
                                                            elif var.getAttribute('name') == 'error_description':
                                                                msg = var.getAttribute('value')
                                                return code, msg
                                            
                        pass
                    elif status == '1':
                        return -102, 'module scheduled for execution'
                    elif status == '2':
                        return -101, 'module is being executed'
                    elif status == '3':
                        return -104, 'failed to locate/execute module'
                    elif status == '5':
                        return -103, 'module removed from schedule'
    
    raise RicciError, str('no ' + str(module_num) + 'th module in the batch, or malformed response')

